class TableSchemeTaskUI {
    constructor(
        nodeParent,     // DOM node to use to display interface
        settings        // Settings beeing individual to the specific task
    ) {
        this._settings = settings;
        this.buildHtml(nodeParent, settings);
    }

    /**
    * Clean up `nodeParent` and build interface into it
    * @return undefined
    */
    buildHtml(
        nodeParent,     // DOM node to use to display interface
        settings        // Settings beeing individual to the specific task
    ) {
        const expectDatatypes = (settings.datatypes instanceof Array) && settings.datatypes.length && (settings.datatypes.length > 1 || settings.datatypes[0]);
        let shuffledDatatypes = [];
        function randomDatatype() {
            if (!expectDatatypes)
                return '';
            if (!shuffledDatatypes.length) {
                shuffledDatatypes = settings.datatypes.filter(s => s).map(s => ({ s, rand: Math.random() }));
                shuffledDatatypes.sort((a, b) => a.rand - b.rand);
                shuffledDatatypes = shuffledDatatypes.map(obj => obj.s);
            }
            return ': ' + shuffledDatatypes.pop();
        }

        function dummyScheme(nr) {
            let s = 'TABELLENNAME' + nr + '(';
            s += '_primaerschlussel' + randomDatatype() + '_, ';
            s += 'nochEinAttribut' + randomDatatype();
            return s + ')';
        }

        nodeParent.textContent = '';
        this._nodeScore = createHtml('div', nodeParent, { 'class': 'punktestand' }, 100);
        this._nodeTask = createHtml('div', nodeParent, { 'class': 'auftrag' });
        this._nodeTask.innerHTML = settings.html;
        this._msgPlace = new ScoTools.MessagePlace();
        this._msgPlace.appendAsChild(nodeParent);
        this._nodeInput = createHtml('textarea', nodeParent, {
            'class': 'tabellenschemainput',
            'placeholder': dummyScheme(settings.expectMultiple ? 1 : '') + (settings.expectMultiple ? ('\n' + dummyScheme(2)) : '')
        });
        this._nodeInput.addEventListener('input', () => {
            if (this.ontextchange instanceof Function)
                this.ontextchange.call(this, this._nodeInput.value);
            this.reportMessage();
        });
        if (expectDatatypes) {
            const nodeDatatypeList = createHtml('div', nodeParent, { 'class': 'datentypen' }, 'Erlaubte Datentypen: ');
            settings.datatypes.forEach((item, index, { length }) => {
                if (index && (index < length - 1))
                    createHtml(false, nodeDatatypeList, null, ', ');
                if (index && (index == length - 1))
                    createHtml(false, nodeDatatypeList, null, ' und ');
                createHtml(item ? 'code' : false, nodeDatatypeList, null, item ? item : 'keine Angabe');
            });
        } else {
            createHtml('div', nodeParent, { 'class': 'datentypen' }, 'Verzichte auf die Angabe von Datentypen');
        }
        createHtml('button', createHtml('div', nodeParent, { 'class': 'centeredBtns' }), { 'type': 'button' }, 'Tabellenschema' + (settings.expectMultiple ? 'ta' : '') + ' überprüfen')
            .addEventListener('click', ev => {
                ev.preventDefault();
                this.check();
            });
    }

    /**
    * Change inputed text to given value
    * @return undefined
    */
    setInput(
        value       // New value
    ) {
        if (value != this._nodeInput.value)
            this._nodeInput.value = value;
    }

    /**
    * Display message above of textarea (or clear it)
    * @return undefined
    */
    reportMessage(
        strText,        // Optional: Text to display
        type            // Optional: Message type (see ScoTools.MSGTYPE_*), defaults to ScoTools.MSGTYPE_FAILED
    ) {
        if (!strText)
            strText = '';
        if (this._lastMessage == strText && this._lastMessageType == type)
            return;
        this._lastMessage = strText;
        this._lastMessageType = type;
        if (!strText)
            this._msgPlace.clear();
        else
            this._msgPlace.add(strText, type);
    }

    /**
    * Check input
    * @return undefined
    */
    check() {
        const tokens = SchemeDefToken.tokenize(this._nodeInput.value);
        // Test syntax
        let defs = null;
        try {
            defs = SchemeDefs.parse(tokens);
        } catch(err) {
            if (err instanceof SchemeDefParseError) {
                if (err.token instanceof SchemeDefToken)
                    this._nodeInput.setSelectionRange(err.token.offset, err.token.offset + err.token.content.length);
                else
                    this._nodeInput.setSelectionRange(this._nodeInput.value.length, this._nodeInput.value.length);
                this._nodeInput.focus();
            }
            this.reportMessage('Syntaxfehler: ' + err.message);
            if (!this.solved)
                this.fine += this._settings.fineSyntax || 0;
            if (this.onaftercheck instanceof Function)
                this.onaftercheck.call(this);
            return;
        }
        // Test semantics
        let datatypes = this._settings.datatypes;
        if (!(datatypes instanceof Array) || !datatypes.length || (datatypes.length == 1 && !datatypes[0]))
            datatypes = undefined;
        const semanticError = defs.validate(datatypes);
        if (semanticError) {
            this.reportMessage('Semantikfehler: ' + semanticError);
            if (!this.solved)
                this.fine += this._settings.fineSemantic || 0;
            if (this.onaftercheck instanceof Function)
                this.onaftercheck.call(this);
            return;
        }
        // Check according to target structure
        const match = defs.match(this._settings.solution)
        if (match.match < 1) {
            this.reportMessage(match.desc || 'Deine Lösung ist fehlerhaft');
            if (!this.solved)
                this.fine += this._settings.fineDiff || 0;
            if (this.onaftercheck instanceof Function)
                this.onaftercheck.call(this);
            return;
        }
        // Report done
        this.solved = true;
        this.reportMessage(this._settings.done || 'Aufgabe gelöst', ScoTools.MSGTYPE_SUCCESS);
        if (this.onaftercheck instanceof Function)
            this.onaftercheck.call(this);
    }

    /**
    * Find out the number of penalty points already collected
    * @return Number    Number of fine points
    */
    getFine() {
        return this._fine;
    }

    /**
    * Set number of penalty points to `value`
    * @return undefined
    */
    setFine(
        value       // New number of fine points
    ) {
        if (!Number.isFinite(value))
            return;
        const newFine = Math.min(100, Math.max(0, Math.round(value)));
        if (newFine == this._fine)
            return;
        this._fine = newFine;
        this._nodeScore.textContent = 100 - this._fine;
    }

    /**
    */
    getState() {
        return {
            fine: this.fine,
            solved: this.solved,
            text: this._nodeInput.value,
            message: this._lastMessage || '',
            messageSuccess: this._lastMessageType == ScoTools.MSGTYPE_SUCCESS
        };
    }
}

// Registered fine count
TableSchemeTaskUI.prototype._fine = 0;
Object.defineProperty(TableSchemeTaskUI.prototype, 'fine', { enumerable: true, set(v) { this.setFine(v); }, get() { return this.getFine(); } });

// Is task allready solved?
TableSchemeTaskUI.prototype.solved = false;

// Function to call after each change of inputed text
TableSchemeTaskUI.prototype.ontextchange = null;

// Function to call after each check of inputed text and eventual change of fine count
TableSchemeTaskUI.prototype.onaftercheck = null;
